/*
 * value.h
 * Copyright 2005 Linuon Inc. Tokyo Japan All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Junji Kanemaru <junji.kanemaru@linuon.com>
 */
#ifndef VALUE_H
#define VALUE_H
#include <string>

/**
** Base Value class
**
** This class is a abstracted container that hides type kind.
** For instance, if you want to have a stack class which can be pushed
** any kind of data like int, void*, std::string, etc. then
** template won't help, it allows only same type.
** We needed to have totally abstracted class to manage abstracted type.
** This class has field "ValueData* _data", yes it is pointer, and
** Value objects share it. Value objects could refer same ValueData object.
** It prevents redundant object creation and copying. It is OK because
** Vlaue obejct is actually container, mostly used as read-only object.
** It is ugly though. I take the blame... -- Junji
*/
class Value {
public:
    /* Supported Value types */
    typedef enum {
        VOID,
        BOOL,
        INT32,
        STRING,
        ARRAY,
        STRUCT,
        PTR,
        DEAD
    } ValueType;

    /* Error code of this class. Will be used with Fault class */
    typedef enum {
        INVALID_TYPE_CAST = 100,
        INDEX_OUT_OF_BOUND,
        NO_SUCH_KEY_FOUND,
        DEAD_VALUE,
        FATAL
    } ValueError;

    /* Constactors and a destractor */
    Value();
    Value(const Value& v);
    virtual ~Value();

    /**
    ** Operator overloadings that abstract real data type of 
    ** derived Value class */
    operator bool() const { return getBool(); }
    operator int() const { return getInt(); };
    operator const char*() const { return getCString(); }
    operator const void*() const { return getPtr(); }
    operator ValueType() const { return _type; }
    virtual operator size_t() const { return size(); }

    /* Interface methods correspond to above operators */
    bool getBool() const;
    int getInt() const;
    const char* getCString() const;
    std::string getString() const;
    const void* getPtr() const;
    ValueType type() const;
    virtual size_t size() const;

    /* Makes a reference to source Value */
    Value& operator =(const Value& v);

    /* Test type */
    bool operator ==(const ValueType& type) const { return (_type == type); }
    /* Test Value */
    bool operator ==(const Value& v) const;

    /* Operator overloadings for ARRAY type */
    Value& operator [](const int idx) const;

    /* Operator overloadings for STRUCT type */
    Value& operator [](const std::string& key) const;
    Value& operator [](const char* key) const;

    /* Returns type string */
    std::string getTypeString() const;

    /* For de-referencing ValueData, used to empty Vlaue */
    static const Value VOID_VALUE;

protected:
    Value(ValueType type, size_t size, const void* data = NULL);

    void decRef();
    void incRef();
    int refCnt() const;

    /* Forward declaration. Internal use only. */
    class ValueData;
    ValueData& valueData() const { return *_data; }

private:
    ValueType _type;

    /**
    ** Placeholder object. Would be shared with other Value objects. 
    ** It will be deleted when there's no other Value object is referencing
    ** to it.  */
    ValueData* _data;
};


/**
** Followings are derived class of Value class above. 
*/
/*
** ValueBoolean */
class ValueBoolean : public Value {
public:
    ValueBoolean(bool b=false);
    ValueBoolean(const Value& val) : Value(val) {}
    ~ValueBoolean() {}
};

/**
** ValueInt32 */
class ValueInt32 : public Value {
public:
    ValueInt32(int i=0);
    ValueInt32(const Value& val) : Value(val) {}
};

/**
** ValueString */
class ValueString : public Value {
public:
    ValueString(const std::string& str);
    ValueString(const Value& val) : Value(val) {}
    ValueString(const char* str);
};

/**
** ValueArray: Array holds reference of each Value, not a copy */
class ValueArray : public Value {
public:
    ValueArray();
    ValueArray(const Value& val) : Value(val) {}
    void append(const Value& val);
    void remove(const Value& val);
    void operator +=(const Value& val);
    void operator -=(const Value& val);
};

/**
** ValueStruct: As same as Array, Struct holds reference of each Value. */
class ValueStruct : public Value {
public:
    ValueStruct();
    ValueStruct(const Value& val) : Value(val) {}
    void set(const std::string& key, const Value& val);
    void set(const char* key, const Value& val);
    Value& get(const std::string& key);
    Value& get(const char* key);
    void erase(const std::string& key);
    void erase(const char* key);
};

/**
** ValuePtr */
class ValuePtr : public Value {
public:
    ValuePtr(size_t size, const void* ptr);
    ValuePtr(const Value& val) : Value(val) {}
};

#endif /* VALUE_H */

