/* auditd-dispatch.c -- 
 * Copyright 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Steve Grubb <sgrubb@redhat.com>
 *   Junji Kanemaru <junji.kanemaru@linuon.com>
 */

#include "config.h"
#include <unistd.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "libaudit.h"
#include "private.h"
#include "auditd-dispatch.h"

/* This is the communications channel between auditd & the dispatcher */
static int pipes[2];
static pid_t pid;

/* set_flags: to set flags to file desc */
static int set_flags(int fn, int flags)
{
	int fl;

	if ((fl = fcntl(fn, F_GETFL, 0)) < 0) {
		audit_msg(LOG_ERR, "fcntl failed. cannot get flags (%s)\n", 
			strerror(errno));
		 return fl;
	}

	fl |= flags;

	return fcntl(fn, F_SETFL, fl);
}

/* This function returns 1 on error & 0 on success */
int init_dispatcher(const struct daemon_conf *config)
{
	if (config->dispatcher == NULL) {
		pipes[0] = -1;
		pipes[1] = -1;
		return 0;
	}

	if (pipe(pipes)) {
		audit_msg(LOG_ERR, "Failed creating pipes");
		return 1;
	}

	/* Make both pipes non-blocking */
	if (set_flags(pipes[0], O_NONBLOCK) < 0) {
		audit_msg(LOG_ERR, "Failed to set O_NONBLOCK flag");
		return 1;
	}
	if (set_flags(pipes[1], O_NONBLOCK) < 0) {
		audit_msg(LOG_ERR, "Failed to set O_NONBLOCK flag");
		return 1;
	}

	// do the fork
	pid = fork();
	switch(pid) {
		case 0:	// child
			close(0);
			dup(pipes[0]);
			close(pipes[0]);
			close(pipes[1]);
			setsid();
			execl(config->dispatcher, config->dispatcher, NULL);
			audit_msg(LOG_ERR, "exec() failed");
			exit(1);
			break;
		case -1:	// error
			return 1;
			break;
		default:	// parent
			close(pipes[0]);
			break;
	}

	return 0;
}

void shutdown_dispatcher(void)
{
	// kill child
	kill(pid, SIGTERM);
	// wait for term
	// if not in time, send sigkill
}

void dispatch_event(const struct audit_reply *rep)
{
	int rc;
	struct iovec vec[2];
	union audit_dispatcher_header hdr;

	hdr.type = rep->type;
	hdr.size = rep->len;

	if (pipes[1] != -1) {
		vec[0].iov_base = (void*)&hdr;
		vec[0].iov_len = sizeof(hdr);
		vec[1].iov_base = (void*)rep->message;
		vec[1].iov_len = rep->len;

		rc = writev(pipes[1], vec, 2);
		// close pipe if no child or peer has been lost
		if (rc <= 0) {
			if (errno == EPIPE)
				pipes[1] = -1;
			else {
				audit_msg(LOG_ERR, "dispatch err: %s",
					strerror(errno));
			}
		}
	}
}

